require "zenoss_api/version"

module ZenossApi
  require 'json'
  require 'httparty'
 
  ROUTERS = { 
    :MessagingRouter => 'messaging',
    :EventsRouter => 'evconsole',
    :ProcessRouter => 'process',
    :ServiceRouter => 'service',
    :DeviceRouter => 'device',
    :NetworkRouter => 'network',
    :TemplateRouter => 'template',
    :DetailNavRouter => 'detailnav',
    :ReportRouter => 'report',
    :MibRouter => 'mib',
    :ZenPackRouter => 'zenpack',
    :TriggersRouter => 'triggers',
    :SettingsRouter => 'settings',
    :JobsRouter => 'jobs',
    :IntrospectionRouter => 'introspection',
    :UsersRouter => 'users',
    :EventClassesRouter => 'Events/evclasses',
  }

  ROUTER_RETURN_KEYS = {
    'DeviceRouter' => 'devices',
    'EventsRouter' => 'events',
  }

  ROUTER_RETURN_TYPE = {
    'TemplateRouter' => 'templates',
  }
  
  DEFAULT_API_METHOD_ARGS = { 
    :DeviceRouter => {
      :getDevices => {
        'uid' => nil,
        'start' => 0,
        'params' => nil,
        'limit' => 50,
        'sort' => 'name',
        'page' => nil,
        'dir' => 'ASC',
        'keys' => nil,
      },
      :getInfo => {
         'uid' => nil,
         'keys' => nil,
      },
    },
    :EventsRouter => {
      :query => {
        'limit' => 50,
      },
    },
  }

  EVENT_STATE = {
    :new          => 0,
    :acknowledged => 1,
    :suppressed   => 2,
    :closed       => 3,
    :cleared      => 4,
    :aged         => 6,
 
  }

  PRODUCTION_STATE = {
    :production     => 1000,
    :pre_production => 500,
    :test           => 400,
    :maintenance    => 300,
    :decomissioned  => -1,
  }

  EVENT_SEVERITY = {
    :clear    => 0, # Green
    :debug    => 1, # Grey
    :info     => 2, # Blue
    :warning  => 3, # Yellow
    :error    => 4, # Orange
    :critical => 5, # Red
  }
  
  class Client
    
    def initialize(url, username, password, verify_ssl = true)
      @url = url
      @auth = {:username => username, :password => password}
      @rec_count = 1
      @headers = {
        'Content-Type' => 'application/json', 
        'Accept' => 'application/json',
      }
      @verify = verify_ssl
    end

    def _paginated_post(url, payload, xoffset=0)
      Enumerator.new do |y|
        total = 0
        offset = 0
        key = ROUTER_RETURN_KEYS[payload[0]['action']]
        limit = payload[0]['data'][0]['limit']
        loop do
          response = self._post(url, payload.to_json)
          totalCount = response['totalCount']
          data = response[key]
          total += data.length
          offset += limit
          payload[0]['data'][0]['start'] = offset
          
          data.each do |element|
            y.yield element
          end

          break if (data.empty? || total >= totalCount)
        end
      end
    end

    # _normalize_response
    # @param [Hash] zenoss router response
    # @return [Hash] normalized response
    def _normalize_response(r)
      data = {
        'action' => r['action'],
        'method' => r['method'],
      }
      result = r['result']
      begin
        data.merge!(result)
      rescue TypeError
        # zenoss returns inconsistent result types, this is not a hash
        type = ROUTER_RETURN_TYPE[data['action']]
        data[type] = result
        data['totalCount'] = result.length
        data['success'] = true
      end 
      data
    end
        
    # http post to api
    # 
    # @param [String] url of api endpoint
    # @param [String] json data
    # @return [Hash] result of api query 
    def _post(url, payload) 
     #### TODO wrap in rescue and handle exceptions
     response = HTTParty.post(url,
       {
         :body => payload,
         :basic_auth => @auth,
         :headers => @headers,
         :timeout => 300,
         :verify => @verify,
       }) 
     if response.code == 200
       _normalize_response(response)
     else 
       raise ({ code: response.code, response: response }).to_json
     end
    end
   
    # api endpoint router method
    # 
    # @param router [String] name of the zenoss api router
    # @param method [String] name of the zenoss api method
    # @param data [hash] api method arguments
    # @param options [Hash] api execution control options
    # @return [Hash] result of api query
    def router(router, method, data={}, options={})
      if data.empty? 
        _data = []
      else 
        _data = [ data ]
      end

      raise "Error: #{router} is not a valid route." unless ROUTERS.key?(router.to_sym)

      api_call = "#{@url}/zport/dmd/#{ROUTERS[router.to_sym]}_router"
      payload = [{ 
        'action' => router,
        'method' => method,
        'data' => _data,
        'type' => 'rpc',
        'tid' => @rec_count
      }]
    
       if options[:paginate]  
         self._paginated_post(api_call, payload)
       else 
         self._post(api_call, payload.to_json) 
       end
    end
 
    def validateUids(uids, type="devices")
      if uids.kind_of?(Array)
        uids.collect { |x| self._validateUid(x, type) }
      else 
        [ self._validateUid(uids, type) ]
      end 
    end
 
    def _validateUid(u, type)
      if u.match(/^\/zport\/dmd\/.*$/)
        u 
      else 
        response = self.getDeviceProperties(u, [ "uid" ])
        if not response.nil?
          response[type][0]['uid']
        else 
          nil
        end
      end
    end 
    
    def getDeviceUidByName(device)
    end

    # set device location
    #
    # @param uids [Array] list of device uids 
    def setDeviceLocation(uids, location)
      validated = self.validateUids(uids)
      data = { 'uids' => uids, 'ranges' => [], 'target' => location }
      response = self.router("DeviceRouter", "moveDevices", data)
    end

    # get device properties
    #
    # @param device [String] the name of the device (device title or id)
    # @param properties [Array] properties to be returned; default uid 
    # @return [Hash] of device properties or nil if device not found 
    def getDeviceProperties(device, properties=['uid'])
      data = { 'keys' => properties, 'params' => { 'name' => device }, }
      result = self.router("DeviceRouter", "getDevices", data)
      if result['success'] == true and result['totalCount'] == 1
        result
      else
        nil
      end
    end

    # helper method to extract uid from a router device query
    #
    # @param result [Hash] response from router request
    # @return [String] device uid
    def getResultUid(result)
      result['devices'][0]['uid']
    end

    # helper method to extract hash from a router query
    #
    # @param result [Hash] response from router request
    # @return [String] device uid
    def getResultHash(result)
      result['hash']
    end
 
    # sets the production state on a device
    #
    # @param device [String] the name of the device (device title or id)
    # @param state [Int] production state to set 
    def setProductionState(device, state = PRODUCTION_STATE[:production])
      result = self.getDeviceProperties(device, ["uid"])
      data = { 
        'uids' => [ self.getResultUid(result) ],
        'hashcheck' => self.getResultHash(result),
        'prodState' => state,
      }
      self.router("DeviceRouter", "setProductionState", data)
    end

    # get a list of devices under a device class
    #
    # @param device_class [String] device class to search
    # @return [hash] 
    def getDevicesInDeviceClass(device_class='/zport/dmd/Devices', keys=['uid'], limit=nil)
      data = { 'uid' => device_class, 'keys' => keys, 'limit' => limit }
      result = self.router('DeviceRouter', 'getDevices', data)
    end
  end
end

